#ifndef _DPT_H
#define _DPT_H

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

#include "sys_info.h"

/*
 * Function Prototypes
 */

int dpt_proc_info(char *buffer, char **start, off_t offset,
                       int length, int inode, int inout);
int dpt_detect(Scsi_Host_Template *sht);
int dpt_queue(Scsi_Cmnd *cmd, void (*cmdcomplete)(Scsi_Cmnd *));
int dpt_abort(Scsi_Cmnd *cmd);
int dpt_reset(Scsi_Cmnd *cmd, unsigned int flags);
int dpt_release(struct Scsi_Host *host);

const char * dpt_info(struct Scsi_Host *pSHost);
int dpt_biosparam(Disk *disk, kdev_t dev, int geom[]);

/*
 * Scsi_Host_Template (see hosts.h) for DPT - some fields
 * to do with card config are filled in after the card is detected.
 */

#define DPT_DRIVER_NAME	"DPT SmartRAID V"

#define DPT_H_VERSION   "1.14"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,101) 	/* see eata.h */

#define DPT_I2O {							\
	NULL,		/* next: next */			\
	NULL,		/* module: usage count */		\
	NULL,		/* proc_dir: ptr to /proc/scsi entry */ 	\
	dpt_proc_info,	/* proc-fs info function */ 	\
	DPT_DRIVER_NAME,/* name: driver name */ 		\
	dpt_detect,	/* detect: Detect host adapter */ 	\
	dpt_release,	/* release: Release host adapter */ 	\
	dpt_info,	/* info: Driver info function */	\
	NULL,		/* command: Command function */		\
	dpt_queue,	/* queuecommand: Queue command function */	\
	dpt_abort,	/* abort: Abort command function */	\
	dpt_reset,	/* reset: Reset command function */	\
	NULL,		/* slave_attach: Slave attach functon */	\
	dpt_biosparam,	/* bios_param: Disk BIOS parameters */	\
	1,		/* can_queue: max # of simultaneous commands */ \
	7,		/* this_id: HBA target id */		\
	0,		/* sg_tablesize: size of scatter gather list */ \
	1,		/* cmd_per_lun: tagged queueing - 1 not supported */ \
	0,		/* present: # of HBA's present (DPT cards) */ \
	0,		/* unchecked_isa_dma: does not use DMA on ISA bus */ \
	ENABLE_CLUSTERING 	/* use_clustering: Enable clustering */	\
}
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,2,16)
#define DPT_I2O {							\
	NULL,		/* next: next */			\
	NULL,		/* module: for loadable modules */	\
	NULL,		/* proc_dir: ptr to /proc/scsi entry */ 	\
	dpt_proc_info,	/* proc_info: proc-fs info function */ 	\
	DPT_DRIVER_NAME,/* name: driver name */ 		\
	dpt_detect,	/* detect: Detect host adapter */ 	\
	dpt_release,	/* release: Release host adapter */ 	\
	dpt_info,	/* info: Driver info function */	\
	NULL,		/* ioctl: ioctl interface */		\
	NULL,		/* command: Command function */		\
	dpt_queue,	/* queuecommand: Queue command function */	\
	NULL,		/* eh_strategy_handler */		\
	NULL,		/* eh_abort_handler */		\
	NULL,		/* eh_device_reset_handler */		\
	NULL,		/* eh_bus_reset_handler */		\
	NULL,		/* eh_host_reset_handler */		\
	dpt_abort,	/* abort: Abort command function */	\
	dpt_reset,	/* reset: (is obselet) Reset command function */ \
	NULL,		/* slave_attach: Slave attach functon */	\
	dpt_biosparam,	/* bios_param: Disk BIOS parameters */	\
	1,		/* can_queue: max # of simultaneous commands */ \
	7,		/* this_id: HBA target id */		\
	0,		/* sg_tablesize: size of scatter gather list */ \
	1,		/* cmd_per_lun: tagged queueing - 1 not supported */ \
	0,		/* present: # of HBA's present (DPT cards) */ \
	0,		/* unchecked_isa_dma: does not use DMA on ISA bus */ \
	ENABLE_CLUSTERING, 	/* use_clustering: Enable clustering */	\
	0,		/* use_new_eh_code: */ \
	0		/* emulated: */ \
}

#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,3,00)
#define DPT_I2O { \
	next: NULL,							\
	module: NULL,							\
	proc_info: dpt_proc_info,					\
	name: DPT_DRIVER_NAME,						\
	detect: dpt_detect,						\
	revoke: NULL,							\
	release: dpt_release,						\
	info: dpt_info,							\
	command: NULL,							\
	queuecommand: dpt_queue,					\
	eh_strategy_handler: NULL,					\
	eh_abort_handler: NULL,						\
	eh_device_reset_handler: NULL,					\
	eh_bus_reset_handler: NULL,					\
	eh_host_reset_handler: NULL,					\
	abort: dpt_abort,						\
	reset: dpt_reset,						\
	slave_attach: NULL,						\
	bios_param: dpt_biosparam,					\
	can_queue: 1,			/* max simultaneous cmds      */\
	this_id: 7,			/* scsi id of host adapter    */\
	sg_tablesize: 0,		/* max scatter-gather cmds    */\
	cmd_per_lun: 1,			/* cmds per lun (linked cmds) */\
	present: 0,			/* number of 7xxx's present   */\
	unchecked_isa_dma: 0,		/* no memory DMA restrictions */\
	use_clustering: ENABLE_CLUSTERING,				\
	use_new_eh_code: 0						\
}

#else /* KERNEL_VERSION > 2.3.00 */

#define DPT_I2O { \
	next: NULL,							\
	module: NULL,							\
	proc_info: dpt_proc_info,					\
	name: DPT_DRIVER_NAME,						\
	detect: dpt_detect,						\
	revoke: NULL,							\
	release: dpt_release,						\
	info: dpt_info,							\
	command: NULL,							\
	queuecommand: dpt_queue,					\
	eh_strategy_handler: NULL,					\
	eh_abort_handler: NULL,						\
	eh_device_reset_handler: NULL,					\
	eh_bus_reset_handler: NULL,					\
	eh_host_reset_handler: NULL,					\
	abort: dpt_abort,						\
	reset: dpt_reset,						\
	slave_attach: NULL,						\
	bios_param: dpt_biosparam,					\
	can_queue: 1,			/* max simultaneous cmds      */\
	this_id: 7,			/* scsi id of host adapter    */\
	sg_tablesize: 0,		/* max scatter-gather cmds    */\
	cmd_per_lun: 1,			/* cmds per lun (linked cmds) */\
	present: 0,			/* number of 7xxx's present   */\
	unchecked_isa_dma: 0,		/* no memory DMA restrictions */\
	use_clustering: ENABLE_CLUSTERING,				\
	use_new_eh_code: 0,						\
	proc_name: "dpt_i2o"						\
}
#endif	/* LINUX_VERSION_CODE > KERNEL_VERSION(2,3,00) */

#ifndef HOSTS_C
# ifdef DEBUG
#  define STATIC
# else
#  define STATIC static
# endif //#ifdef DEBUG


//----------------------------------------------------------------------------
//                        Miscellaneous Definitions                          
//----------------------------------------------------------------------------

//
// SCSI Definitions 
//
#define FLUSH_CACHE_CMD       0x35        // SCSI Synchronize Cache Command  

//
// Configuration Definitions 
//
#define HEADS                 64          // Number Of Heads For Disks.      
#define SECTORS               64          // Number Of SPT For Disks.        
#define SG_LIST_ELEMENTS      64          // Number Scatter Gather Elements
#define R_LIMIT               0x20000L    // Spin Loop Limit                 
#define SCRATCH_BUFFER_SIZE   1024 * 64   // Scratch buffer size in bytes
#define MSG_ZERO_SIZE         24          // I2O Message Pachet Clear Size
#define PRINT_BUFFER_SIZE     512         // Debugging print buffer size
   
/*
 * NOTE:  In the 2.0.xx linux tree, setting MAX_TO_IOP_MESSAGES to 220
 * will cause the driver to break.  I believe you can set it higher than
 * 210 in the 2.2.xx tree.  You will have to test it.
 */
#define MAX_TO_IOP_MESSAGES   210          // Max Inbound Status Packets      
#define MAX_FROM_IOP_MESSAGES 64          // Max Outbound Message Packets    

#define TO_IOP_FIFO_OFFSET    0x40        // Byte offset for to IOP fifo
#define FROM_IOP_FIFO_OFFSET  0x44        // Byte offset for from IOP fifo
#define INT_STATUS_OFFSET     0x30        // Byte offset for int status reg 
#define INT_MASK_OFFSET       0x34        // Byte offset for int mask reg 
#define WAIT_FOR_IOP_RESET    5000000     // IOP Reset timeout value
#define MAX_GET_STATUS_RETRY  20000000    // Number Of Retries for Get Status
#define WAIT_FOR_GET_STATUS   5000

#define EMPTY_QUEUE           0xffffffff  // return value for an enpty queue 

#define MAX_TARGET_ID         128         // Maximum Target ID Supported     
#define MAX_LUN               16          // Maximum LUN Supported           
#define MAX_CHANNEL           5           // Maximum Channel # Supported     
#define MAX_USER_SG           8           // Maximum # SG entries for IOCTL  

#define MAXADAPTERS           4          // Maximum # HBAs We Support       
#define MAXDEV                56          // Maximum # Devices We Support    

#ifndef TRUE
#define TRUE                  1
#define FALSE                 0
#endif

//
// Timeout Running Frequency. 
//

#define TIMEOUT_FREQ          (30 * Hz)          
#define CMD_TIMEOUT_FREQ      (20 * Hz) + lbolt

//typedef REQ_IO                OS_REQUEST_T   // OS Request Packet
//typedef REQ_IO                *pOS_REQUEST_T // OS Request Packet

#define OS_REQUEST_T          REQ_IO // OS Request Packet
#define pOS_REQUEST_T         REQ_IO * // OS Request Packet

//
// Virtual HBA Entry Definition
//
typedef struct OS_HBA
{
  INT32 HbaNum;
  INT32 BusNum;

} OS_HBA_T;

//
// Adapter HbaBusType Field Definitions 
//

#define DPT_BUS_ISA    0
#define DPT_BUS_EISA   1
#define DPT_BUS_PCI    2

//----------------------------------------------------------------------------
//                            PCI Definitions
//----------------------------------------------------------------------------

//
// Configuration Space Register Offset Definitions 
//

#define PCI_REG_VENDOR_ID         0x00     // Vendor ID                       
#define PCI_REG_DEVICE_ID         0x02     // Device ID                       
#define PCI_REG_COMMAND           0x04     // Command                         
#define PCI_REG_STATUS            0x06     // Status                          
#define PCI_REG_REVISION_ID       0x08     // Revision ID                     
#define PCI_REG_CLASS_CODE        0x09     // Class Code                      
#define PCI_REG_CACHE_LINE_SIZE   0x0c     // Cache Line Size                 
#define PCI_REG_LATENCY_TIMER     0x0d     // Latency Timer                   
#define PCI_REG_HEADER_TYPE       0x0e     // Header type                     
#define PCI_REG_BIST              0x0f     // BIST                            
#define PCI_REG_BASE_ADDRESS      0x10     // First Base Address Register     
#define PCI_REG_ROM_ADDRESS       0x30     // ROM Address                     
#define PCI_REG_IRQ_lINE          0x3c     // Interrupt line                  
#define PCI_REG_IRQ_PIN           0x3d     // Interrupt Pin                   
#define PCI_REG_MIN_GNT           0x3e     // Min_Gnt                         
#define PCI_REG_MAX_LAT           0x3f     // Max_Lat                         

//
// DPT Specific Configuration Space Register Offset Definitions 
//

#define PCI_REG_DPT_CONFIG        0x40     //                                 
#define PCI_REG_PUMP_MODE_ADDRESS 0x44     //                                 
#define PCI_REG_PUMP_MODE_DATA    0x48     //                                 
#define PCI_REG_CONFIG_PARAM1     0x50     //                                 
#define PCI_REG_CONFIG_PARAM2     0x54     //                                 

//
// DPT PCI Specific Definitions 
//

#define PCI_DPT_VENDOR_ID         0x1044   // DPT PCI Vendor ID               
#define PCI_DPT_DEVICE_ID         0xA501   // DPT PCI I2O Device ID
#define PCI_DPT_BASE_CLASS        0x05     // PCI Memory Controller Base Class
#define PCI_DPT_SUB_CLASS         0x80     // PCI Other Sub Class      

//
// Command Register Bit Field Definitions 
//

#define PCI_CMD_IO_ENABLED        0x01     // I/O Space Enabled               
#define PCI_CMD_MEMORY_ENABLED    0x02     // Memory Space Enabled
#define PCI_CMD_BUSMASTER         0x04     // Bus Master Enabled              


#ifndef EVEREST

//
// PCI Hardware Port Definitions For Configuration Method 1 
//

#define CONFIG_ADDRESS_REG        0x0CF8   // Config Address Register         
#define CONFIG_DATA_REG           0x0CFC   // Config Data Access Register     

//
// PCI Hardware Port Definitions For Configuration Method 2 
//

#define CSE_REG                   0x0CF8   // Config Space Enable             
#define FORWARD_REG               0x0CFA   // BUS Access Register             
#define CONFIG_SPACE_REG          0xC000   // Start Of Config Space Mapping   

//
// PCI Bus And Device Limitations 
//

#define MAX_PCI_DEVICES           32       // Maximum # Of Devices Per Bus   
#define MAX_METHOD_2              16       // Max Devices For Method 2       
#define MAX_PCI_BUS               16       // Maximum # Of Busses Allowed    
#define MAX_PCI_FUNCTIONS         8        // Maximum # Of Functions

//
// PCI Configuration Space Structure 
//

typedef struct PCI_CONFIG_DATA
{
  UINT16 VendorID;            // Vendor ID                      
  UINT16 DeviceID;            // Device ID                      
  UINT16 Command;             // Command Register               
  UINT16 Status;              // Status Register                
  UINT8  RevisionID;          // Revision ID                    
  UINT8  ClassProg;           // Reg. Level Programing Interface
  UINT8  SubClass;            // Device Sub Class               
  UINT8  BaseClass;           // Device Base Class              
  UINT8  CacheLineSize;       // Cache Line Size                
  UINT8  LatencyTimer;        // Latency Timer                  
  UINT8  HeaderType;          // Header Type                    
  UINT8  BIST;                // Built In Self Test             
  UINT32 BaseAddress0;        // Base Address Register 0        
  UINT32 BaseAddress1;        // Base Address Register 1        
  UINT32 BaseAddress2;        // Base Address Register 2        
  UINT32 BaseAddress3;        // Base Address Register 3        
  UINT32 BaseAddress4;        // Base Address Register 4        
  UINT32 BaseAddress5;        // Base Address Register 5        
  UINT32 Reserved0[2];        // Reserved                       
  UINT32 RomAddress;          // ROM Address                    
  UINT32 Reserved1[2];        // Reserved                       
  UINT8  InterruptLine;       // Interrupt Line Routing Info.   
  UINT8  InterruptPin;        // Interrupt Pin Used             
  UINT8  Min_Gnt;             // Minimum Burst Period Needed    
  UINT8  Max_Lat;             // How Often PCI Bus Access       
 
} PCI_CONFIG_DATA_T;

//
// PCI device list entry definition for the PCI calls
//

typedef struct PCI_DEVICE_INFO
{
  UINT8  DeviceNumber;           // Device Number                 
  UINT8  FunctionNumber;         // Function Number                 
  UINT8  BusNumber;              // Device Bus Number             

} PCI_DEVICE_INFO_T;

//
// SCO UNIX PCI Structure Definitions 
//

struct pci_devinfo 
{
  UINT16 slotnum;                 // PCI Device Number              
  UINT16 funcnum;                 // PCI Device Function Number      
  UINT16 busnum;                  // PCI Bus number                 
};

struct pci_businfo 
{
  UINT16 numbuses;                // Number Of PCI Buses            
  UINT16 mechanism;
};

#endif //#ifndef EVEREST



//
// DPT I2O Device List Entry Definition
//

typedef struct I2O_DEVICE 
{
  struct  pci_devinfo PciInfo;              // PCI Device Info
  UINT32  PhysicalBaseAddress;              // Physical Base Address
  UINT8   *VirtualBaseAddress;              // Physical Base Address
  UINT32  MemorySizeInBytes;                // Size In Bytes 
  UINT32  InterruptVector;                  // Interrupt Vector used
  I2O_EXEC_STATUS_GET_REPLY GetStatusInfo;  // Get Status Data

} I2O_DEVICE_T, *pI2O_DEVICE_T;

//
// SCSI Device Structure Definition 
//

typedef struct DEVINFO
{
/*  UINT32  HbaNum;                 // Host Adapter # For This Device   */
  UINT32  ChannelNum;             // Host Adapter Channel Number     
  UINT32  ScsiTargetID;           // SCSI Target ID For This Device   
  UINT32  ScsiLUN;                // SCSI LUN For This Device        
  UINT32  TID;                    // I2O ID Of The Device
  UINT32  Flags;                  // Flags For This Device           
  UINT32  DevType;                // SCSI Device Type                
  UINT32  Capacity;               // Number Of Blocks On This Dev    
  UINT32  BlockSize;              // Bytes Per Block For This Dev    
  struct  DEVINFO *NextDevice_P;  // Next Pointer For Dev Table      
  struct  DEVINFO *NextLun_P;     // Next Pointer For Dev Table      
  Scsi_Device *ScsiDev_P;         // Keep a pointer to the device in SCSI layer

} DEVINFO_T, *pDEVINFO_T;

#define DEVINFO_FLAGS_SCSI_PORT_B      0x0001  // SCSI Port Class
#define DEVINFO_FLAGS_SCSI_DEVICE_B    0x0002  // SCSI Device Class
#define DEVINFO_FLAGS_BLOCK_DEVICE_B   0x0004  // Block Device Class
#define DEV_IN_RESET			0x0008
#define DEVINFO_FLAGS_UNSCANNED   0x0010

#define SG_LIST_SIZE  (sizeof(I2O_SG_ELEMENT) * SG_LIST_ELEMENTS)
#define MESSAGE_SIZE  sizeof(PRIVATE_SCSI_SCB_EXECUTE_MESSAGE)
#define MAX_MESSAGE_SIZE  (MESSAGE_SIZE + SG_LIST_SIZE)
#define USER_SG_LIST_SIZE  (sizeof(I2O_SG_ELEMENT) * MAX_USER_SG)
#define MAX_USER_MESSAGE_SIZE (MESSAGE_SIZE + USER_SG_LIST_SIZE)
//
// DPT Message Packet Definition
//

typedef struct DPT_MSG
{
  UINT32         TimeoutValue;               // Command Timeout Value           
  volatile UINT32         Flags;                      // Control Flags Field          
	Scsi_Cmnd	*scsicmd;		/* passed in scsi command */
  struct DPT_MSG *Next_P;                    // Pointer To Next CCB In List.    
  struct DPT_MSG *Previous_P;                // Pointer To Previous CCB In List 
  INT32          (*MSG_Post_P)(struct DPT_MSG *);            // Command Complete Post Routine   
  volatile UINT32         CommandComplete;            // IOCTL Complete Flag
  volatile I2O_INITIATOR_CONTEXT unique;	/* unique ID for this message */
  UINT32         HbaNum;                     // Hba Number
  pDEVINFO_T     DevInfo_P;                  // Device associated with command
  UINT32         ToIopMsgOffset;             // Offset to the I2O message
  volatile UINT8          *ToIopMsg_P;                // Actual Messge Pointer
  UINT8          *LocalToIopMsg;   // ptr. to Local Message Copy
#ifdef LOCALREPLY
  PI2O_SCSI_ERROR_REPLY_MESSAGE_FRAME Reply;  // Ptr. to Reply Message Copy
#else
  I2O_SCSI_ERROR_REPLY_MESSAGE_FRAME Reply;  // Reply Message Copy
#endif	/* LOCALREPLY */
  UINT8          CommandCompleted;			// Used by pass through 

} DPT_MSG_T, *pDPT_MSG_T;


//
// Msg Flags Field Bit Definitions 
//

#define MSG_FLAG_TIMEOUT_B  0x00000001          // Command Timmed Out Flag      
#define MSG_FLAG_RESET_B    0x00000002          // Command Timmed Out And Reset 
#define MSG_FLAG_ACTIVE_B   0x00000004          // Message Active              
#define MSG_FLAG_IOCTL_B    0x00000008          // IOCTL Message 
#define MSG_FLAG_IS_RESET   0x00000010          // is a reset message
#define MSG_FLAG_INQUIRY   0x00000020          // inquire msg is outstanding

typedef struct dpt_queue {
        volatile void *head;
        volatile void *tail;
} dptqueue;

/*
 * doneq is a queue of commands completed by the interrupt handler
 * sendq is a queue of commands that need to be sent to the board
 * resetq is a queue of commands that need to be sent to the board
 *	when the current reset has completed
 */

//#include <asm/spinlock.h>
//
// DPT I2O HBA Structure Definition
//
typedef struct DPT_HBA
{
	struct DPT_HBA		*next;	/* linked list of DPT cards */
	struct	Scsi_Host	*host;	/* ptr to scsi_host struct for card */
	int			cardidx;	/* index into dpti20card table */
	int			host_no;	/* SCSI host number */
	UINT8			inisr;	/* positive if in interrupt routine */
					/* their HbaNum */
  volatile UINT32        Flags;            // Operational State Flags         
	UINT32	maxfromiopmsgs;
	char	name[50];		/* name of card */
  DEVINFO_T     ChannelInfo[4];   // Device Info For Each Channel
  pI2O_DEVICE_T I2oDeviceInfo_P;  // I2O Device Info 
  pDEVINFO_T    DeviceList_P;     // Device list 
  pDEVINFO_T    ScsiDeviceTable_P[MAX_CHANNEL][MAX_TARGET_ID]; // Device Table
  UINT8         *VirtualBaseAddress_P; // Virtual Base Address Of IOP     
  volatile UINT32        *ToIopFifo_P;     // Fifo for messages to the IOP    
  volatile UINT32        *FromIopFifo_P;   // Fifo for messages from the IOP  
  volatile UINT32        *IntStatusReg_P;  // 
  volatile UINT32        *IntMaskReg_P;    // 
  UINT32        OutstandingMsgs;  // # Of CCBs Outstanding To HBA    
  UINT32        QueueDepth;       // Controller Queue depth.         
  UINT32        SgListSize;       // Scatter/Gather List Size.       
  UINT32        MaxChannel;       // Maximum Channel Number          
  UINT32        MaxID;            // Maximum Target ID               
  UINT32        MaxLUN;           // Maximum LUN                     
  INT32         WakeBase;         // MPX Wakeup Base Index.          
  spinlock_t HardwareLock;	// MPX Hardware Lock Structure
  spinlock_t MsgListLock;	// MPX CCB List Lock Structure
  spinlock_t OsQueueLock;	// MPX OS Queue Lock Structure
  spinlock_t IoctlLock;		// MPX IOCTL Lock Structure
  spinlock_t driverlock;	// MPX IOCTL Lock Structure
  spinlock_t flaglock;		// MPX IOCTL Lock Structure
  spinlock_t doneq_lock;
  spinlock_t sendq_lock;
  spinlock_t resetq_lock;

  volatile pDPT_MSG_T    DptMsgActiveList_P;  // Pointer To Next Active CCB.
  volatile pDPT_MSG_T    DptMsgFreeList_P;    // Pointer To Next Free CCB.
  volatile dptqueue doneq; /* commands completed by the interrupt handler */
  volatile dptqueue sendq; /* commands that need to be sent to the board */
  volatile dptqueue resetq;/* commands to be sent to board after reset done */

#ifdef KORIG
  volatile DPT_MSG_T     DptMsg[MAX_TO_IOP_MESSAGES]; // Message List
#endif
  volatile pDPT_MSG_T    dptmsgs;    // Ptr to kmalloc'ed dptmsgs
  volatile I2O_SCSI_ERROR_REPLY_MESSAGE_FRAME FromIopMsg[MAX_FROM_IOP_MESSAGES];
                                             // Reply List
  volatile pUINT8        FwDebugBuffer_P;        // Virtual Address Of FW Debug Buffer
  volatile pUINT32       FwDebugFlags_P;         // Virtual Address Of FW Debug Flags 
  volatile UINT32        FwDebugFlags;           // FW Debug Flags 
  volatile pUINT32       FwDebugStrLength_P;     // Virtual Addr Of FW Debug String Len
  volatile pUINT8        FwDebugBLEDflag_P;      // Virtual Addr Of FW Debug BLED
  volatile pUINT8        FwDebugBLEDvalue_P;     // Virtual Addr Of FW Debug BLED
  UINT32        FwDebugBufferSize;      // FW Debug Buffer Size In Bytes
  volatile UINT8         PrintBuffer_P[PRINT_BUFFER_SIZE];  // Debugging Print Buffer
} DPT_HBA_T, *pDPT_HBA_T;

//
// HBA Flags Field Bit Definitions 
//

#define HBA_FLAGS_INSTALLED_B       0x00000001  // Adapter Was Installed        
#define HBA_FLAGS_BLINKLED_B        0x00000002  // Adapter In Blink LED State   
#define HBA_FLAGS_CACHE_B           0x00000004  // Adapter Has Cache            

#define DPT_IN_ISR		0x00000008	/* in interrupt handler or bh */
#define DPT_IN_BH		0x00000010	/* in bottom half */
#define DPT_IN_ABORT		0x00000020	/* in abort */
#define HBA_FLAGS_IN_RESET	0x00000040	/* in reset */
#define HBA_HOSTRESET_FAILED	0x00000080	/* dpt_resethost failed */




#define HBA_FLAGS_DBG_FLAGS_MASK         0xffff0000  // Mask for debug flags
#define HBA_FLAGS_DBG_KERNEL_PRINT_B     0x00010000  // Kernel Debugger Print
#define HBA_FLAGS_DBG_FW_PRINT_B         0x00020000  // Firmware Debugger Print
#define HBA_FLAGS_DBG_FUNCTION_ENTRY_B   0x00040000  // Function Entry Point
#define HBA_FLAGS_DBG_FUNCTION_EXIT_B    0x00080000  // Function Exit
#define HBA_FLAGS_DBG_ERROR_B            0x00100000  // Error Conditions
#define HBA_FLAGS_DBG_INIT_B             0x00200000  // Init Prints      
#define HBA_FLAGS_DBG_OS_COMMANDS_B      0x00400000  // OS Command Info  
#define HBA_FLAGS_DBG_SCAN_B             0x00800000  // Device Scan      


#define FW_DEBUG_STR_LENGTH_OFFSET 0
#define FW_DEBUG_FLAGS_OFFSET      4
#define FW_DEBUG_BLED_OFFSET       8

#define FW_DEBUG_FLAGS_NO_HEADERS_B    0x01 

#define I2O_INTERRUPT_PENDING_B     0x08
#define I2O_INTERRUPTS_DISABLED_B   0x08

//
// DPT I2O System Table Structure Definition
//

typedef struct I2O_SYSTEM_TABLE 
{
  I2O_SET_SYSTAB_HEADER Header;
  I2O_IOP_ENTRY SysTabEntry[MAXADAPTERS];

}  I2O_SYSTEM_TABLE_T, *pI2O_SYSTEM_TABLE_T;


#endif /* !HOSTS_C */


#ifdef    __cplusplus
   struct drvrHBAinfo_S {
#else
   typedef struct {
#endif

    uSHORT length;        // Remaining length of this
    uSHORT drvrHBAnum;    // Relative HBA # used by the driver
    uLONG  baseAddr;      // Base I/O address
    uSHORT blinkState;    // Blink LED state (0=Not in blink LED)
    uCHAR  pciBusNum;     // PCI Bus # (Optional)
    uCHAR  pciDeviceNum;  // PCI Device # (Optional)
    uSHORT hbaFlags;      // Miscellaneous HBA flags
    uSHORT Interrupt;     // Interrupt set for this device.
#   if (defined(_DPT_ARC))
        uLONG           baseLength;
        ADAPTER_OBJECT *AdapterObject;
        LARGE_INTEGER   DmaLogicalAddress;
        PVOID           DmaVirtualAddress;
        LARGE_INTEGER   ReplyLogicalAddress;
        PVOID           ReplyVirtualAddress;
#   else
        uLONG  reserved1;     // Reserved for future expansion
        uLONG  reserved2;     // Reserved for future expansion
        uLONG  reserved3;     // Reserved for future expansion
#   endif

#ifdef    __cplusplus

//---------- Portability Additions ----------- in sp_eosd.cpp
#ifdef DPT_PORTABLE
        uSHORT          netInsert(dptBuffer_S *buffer);
        uSHORT          netExtract(dptBuffer_S *buffer);
#endif  // DPT_PORTABLE
//--------------------------------------------

   };
#else
   } drvrHBAinfo_S;
#endif

#endif	/* _DPT_H */
